import bpy


def empty_scene() -> None:
    blender_objects = bpy.data.objects

    for obj in bpy.data.objects:
        blender_objects.remove(obj, do_unlink=True)


def load_mesh_into_object(name: str, mesh) -> None:
    # make object from mesh
    new_object = bpy.data.objects.new(name, mesh)

    # make collection
    new_collection = bpy.data.collections.new("new_collection")
    bpy.context.scene.collection.children.link(new_collection)

    # add object to scene collection
    new_collection.objects.link(new_object)

    return new_object



import bmesh
import bpy
from pydantic_numpy import NDArray, NDArrayFp64

from meshemy.blender.shortcut.io import load_mesh_into_object
from meshemy.blender.shortcut.select import select_object


def load_mesh_from_numpy_arrays(
    vertices: NDArrayFp64, edges: NDArray | None, faces: NDArray | None, name: str = "new_object"
):
    assert edges is not None or faces is not None

    # https://b3d.interplanety.org/en/how-to-create-mesh-through-the-blender-python-api/
    blender_mesh = bpy.data.meshes.new(f"{name}mesh")
    blender_mesh.from_pydata(vertices, () if edges is None else edges, () if faces is None else faces)
    blender_mesh.update()

    return load_mesh_into_object(name, blender_mesh)


def triangular_bmesh(mesh_object_name: str) -> tuple[NDArrayFp64, NDArray, NDArray]:
    _ob = select_object(mesh_object_name)

    bpy.ops.object.mode_set(mode="EDIT")
    bpy.ops.mesh.quads_convert_to_tris(quad_method="BEAUTY")
    bpy.ops.object.mode_set(mode="OBJECT")

    bm = bmesh.new()
    bm.from_mesh(select_object(mesh_object_name).data)

    return bm



from typing import Any, Optional

import bpy
from ordered_set import OrderedSet


def select_object(name: str):
    bpy.ops.object.mode_set(mode="OBJECT")
    ob = bpy.context.scene.objects[name]
    bpy.ops.object.select_all(action="DESELECT")  # Deselect all objects
    bpy.context.view_layer.objects.active = ob  # Make the cube the active object
    ob.select_set(True)
    return ob


def select_one_or_all(name: Optional[str] = None) -> Any:
    if name:
        return select_object(name)
    bpy.ops.mesh.select_all(action="SELECT")


def all_meshes_in_scene() -> OrderedSet:
    return OrderedSet([o for o in bpy.context.scene.objects if o.type == "MESH"])


def latest_mesh():
    return all_meshes_in_scene()[0]



import logging
from typing import Optional

import bpy
import numpy as np
from bmesh.types import BMesh
from pydantic import FilePath
from pydantic_numpy import NDArray, NDArrayFp64

from meshemy.blender.constant import SUFFIX_TO_READER, SUFFIX_TO_WRITER
from meshemy.blender.shortcut.select import latest_mesh, select_object
from meshemy.blender.utils import load_mesh_from_numpy_arrays, triangular_bmesh
from meshemy.blender.workflows import merge_close, planar_decimate_mesh
from meshemy.cookbook.base import BaseCookbook

logger = logging.getLogger(__file__)


class BlenderCookbook(BaseCookbook):
    mesh_name: str

    cached_bm: BMesh | None

    def reset_bm(self) -> None:
        self.cached_bm = None

    @property
    def triangular_bmesh(self) -> BMesh:
        if not self.cached_bm:
            self.cached_bm = triangular_bmesh(mesh_object_name=self.mesh_name)
        return self.cached_bm

    @property
    def vertices_numpy_array(self) -> NDArrayFp64 | None:
        return np.array([v.co.copy() for v in self.triangular_bmesh.verts], dtype=np.float64)

    @property
    def edges_numpy_array(self) -> NDArray | None:
        return np.array([[v.index for v in e.verts] for e in self.triangular_bmesh.edges])

    @property
    def faces_numpy_array(self) -> NDArray | None:
        return np.array([[v.index for v in f.verts] for f in self.triangular_bmesh.faces])

    @property
    def watertight(self) -> bool:
        for edge in self.triangular_bmesh.edges:
            if len(edge.link_faces) != 2:
                return False
        return True

    def planar_decimate(self, degree_tol: float) -> None:
        planar_decimate_mesh(degree_tol, mesh_object_name=self.mesh_name)
        logger.debug(f"Planar decimation on {self.mesh_name}, degree tolerance {degree_tol}")
        self.reset_bm()

    def merge_close(self, distance_tol: float) -> None:
        merge_close(distance_tol, mesh_object_name=self.mesh_name)
        logger.debug(f"Merging proximal vertices on {self.mesh_name}, distance tolerance {distance_tol}")
        self.reset_bm()

    def save(self, path: FilePath) -> None:
        _ob = select_object(self.mesh_name)
        SUFFIX_TO_WRITER[path.suffix](filepath=str(path))

    @classmethod
    def from_data(
        cls, vertices: NDArrayFp64, edges: NDArray | None, faces: NDArray | None, name: str = "new_object"
    ) -> "BlenderCookbook":
        load_mesh_from_numpy_arrays(vertices, edges, faces, name)
        return cls(mesh_name=name)

    @classmethod
    def from_file(cls, path: FilePath, name: Optional[str] = None, **kwargs) -> "BlenderCookbook":
        SUFFIX_TO_READER[path.suffix](filepath=str(path), **kwargs)
        if name:
            ob = latest_mesh()
            ob.name = name
            ob.data.name = name
        else:
            name = latest_mesh().name

        return cls(mesh_name=name)



from typing import Callable, Type


def _raise_not_installed_error_callable(
    exception_class: Type[Exception], name: str, extras_name: str
) -> Callable[[], None]:
    def raise_not_installed_error(e: Exception) -> None:
        raise exception_class(
            f"""
            {name} module, {extras_name}, must be installed.

            Please run `pip install meshemy[{extras_name}]`
            """
        ) from e

    return raise_not_installed_error


class BlenderNotInstalledError(ModuleNotFoundError):
    pass


class Open3DNotInstalledError(ModuleNotFoundError):
    pass


class TrimeshNotInstalledError(ModuleNotFoundError):
    pass


blender_module_not_installed_error = _raise_not_installed_error_callable(BlenderNotInstalledError, "Blender", "bpy")
open3d_module_not_installed_error = _raise_not_installed_error_callable(Open3DNotInstalledError, "Open3D", "open3d")
trimesh_module_not_installed_error = _raise_not_installed_error_callable(TrimeshNotInstalledError, "Trimesh", "trimesh")


def more_than_one_mesh_in_scene(number_of_meshes: int) -> str:
    return f"Meshemy only supports files with one mesh. More than one mesh in current scene, {number_of_meshes} meshes"



from typing import Sequence

try:
    import trimesh
except ModuleNotFoundError as e:
    from meshemy.utility.exception import (
        more_than_one_mesh_in_scene,
        trimesh_module_not_installed_error,
    )

    trimesh_module_not_installed_error(e)

from pathlib import Path

import numpy as np
from pydantic import FilePath
from pydantic_numpy import NDArray, NDArrayFp64

from meshemy.cookbook.base import BaseCookbook, MeshIsObjectMixin


def _load_trimesh(path_to_file: FilePath) -> trimesh.Trimesh:
    result = trimesh.load_mesh(path_to_file)
    if isinstance(result, trimesh.Scene):
        meshes = tuple(result.geometry.values())
        assert (number_of_meshes := len(meshes)) == 1, more_than_one_mesh_in_scene(number_of_meshes)
        result = meshes[0]
    return result


class TrimeshCookbook(BaseCookbook, MeshIsObjectMixin[trimesh.Trimesh]):
    mesh_from_file_loader = _load_trimesh

    @property
    def vertices_numpy_array(self) -> NDArrayFp64 | None:
        return np.array(self.mesh.vertices)

    @property
    def edges_numpy_array(self) -> NDArray | None:
        return np.array(self.mesh.edges)

    @property
    def faces_numpy_array(self) -> NDArray | None:
        return np.array(self.mesh.faces)

    @property
    def watertight(self) -> bool:
        return self.mesh.is_watertight

    def save(self, save_path: Path | str) -> None:
        self.mesh.export(save_path)

    def contains(self, vertices: Sequence[NDArray]) -> bool:
        return np.all(self.mesh.contains(vertices))
